package be.iminds.ilabt.jfed.experimenter_gui.editor.properties.bonfire;

import be.iminds.ilabt.jfed.experimenter_gui.config.AdvertisementFetchManager;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesDialog;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesTab;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.SliverType;
import be.iminds.ilabt.jfed.rspec.model.bonfire.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.bonfire.Lan;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.util.CheckboxList;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.Lists;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.ComboBox;
import javafx.scene.control.ProgressBar;
import javafx.scene.layout.VBox;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.stream.Collectors;

public class BonfireNodePropertiesTab extends AdaptivePropertiesTab<FXRspecNode> {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(BonfireNodePropertiesTab.class);
    private final AdaptivePropertiesDialog adaptivePropertiesDialog;
    private final JFedGuiConfig config;
    private final AdvertisementFetchManager advertisementFetchManager;

    @FXML
    private Node root;
    @FXML
    private ComboBox<String> locationComboBox;
    @FXML
    private ComboBox<String> sliverTypeComboBox;
    @FXML
    private CheckboxList<Lan> lans;
    @FXML
    private CheckboxList<DiskImage> diskImages;
    @FXML
    private ProgressBar networksProgressBar;
    @FXML
    private ProgressBar storagesProgressBar;
    private String lastInitializedLocation = null;
    private String lastLocationForWhichDefaultNetworkSelected = null;
    private String lastLocationForWhichDefaultStorageSelected = null;


    @FXML
    private final ReadOnlyBooleanProperty hasFixedNode;// = new SimpleBooleanProperty(true);
    @FXML
    private VBox noAdvertisementBox;
    @FXML
    private VBox loadingAdvertisementBox;

    private final ObservableList<String> locationIds = FXCollections.observableArrayList();
    private final ObservableList<String> sliverTypes = FXCollections.observableArrayList();

    private FXModelRspec currentAdvertisement;
    private Resource currentResource;
    private BonfireNodeInfo currentLocationBonfireNodeInfo = null;

    public BonfireNodePropertiesTab(AdaptivePropertiesDialog adaptivePropertiesDialog) {
        super("Bonfire");
        assert adaptivePropertiesDialog != null;
        this.adaptivePropertiesDialog = adaptivePropertiesDialog;
        this.config = adaptivePropertiesDialog.getConfig();
        this.advertisementFetchManager = adaptivePropertiesDialog.getAdvertisementFetchManager();
        this.hasFixedNode = adaptivePropertiesDialog.isSpecificNodeChosenProperty();

        //initialize GUI
        setContent(FXMLUtil.createFromFXML(this));

        lans.setStringConverter(StringConverters.BONFIRE_LAN_STRING_CONVERTER);
        diskImages.setStringConverter(StringConverters.BONFIRE_DISK_IMAGE_STRING_CONVERTER);

        sliverTypeComboBox.setItems(sliverTypes);
        locationComboBox.setItems(locationIds);


        //bind to ComponentManagerInfo from AdaptivePropertiesDialog
        assert adaptivePropertiesDialog.selectedResourceProperty() != null;

        ChangeListener<AdvertisementRspecSource> advertisementChangeListener = (observable, oldValue, newValue) -> {
            updateAdvertisement(config.getAdvertisement(currentResource.getServer(), Optional.<Boolean>empty()/*= any advertisement*/).get());
        };

        final ChangeListener<Resource> resourceChangeListener = new ChangeListener<Resource>() {
            @Override
            public void changed(ObservableValue<? extends Resource> observable, Resource oldValue, Resource newValue) {
                if (oldValue != null) {
                    config.getAdvertisement(oldValue.getServer(), Optional.<Boolean>empty()/*= any advertisement*/).
                            removeListener(advertisementChangeListener);
                }
                noAdvertisementBox.visibleProperty().unbind();
                loadingAdvertisementBox.visibleProperty().unbind();

                currentResource = newValue;
                if (newValue != null) {
                    LOG.debug("Now listening for new advertisements of {}", newValue.getServer().getName());
                    ReadOnlyObjectProperty<AdvertisementRspecSource> anyAdvProp =
                            config.getAdvertisement(newValue.getServer(), Optional.<Boolean>empty()/*= any advertisement*/);
                    anyAdvProp.addListener(advertisementChangeListener);
                    noAdvertisementBox.visibleProperty().bind(anyAdvProp.isNull());
                    loadingAdvertisementBox.visibleProperty().bind(
                            advertisementFetchManager.getAllResources(newValue.getServer()).currentRefreshAdvertisementTaskProperty().isNotNull());

                    updateAdvertisement(anyAdvProp.get()); //force initial update
                } else {
                    updateAdvertisement(null);
                }
            }
        };
        adaptivePropertiesDialog.selectedResourceProperty().addListener(resourceChangeListener);
        if (adaptivePropertiesDialog.selectedResourceProperty().get() != null) {
            //already call myself if there is already a selection
            resourceChangeListener.changed(adaptivePropertiesDialog.selectedResourceProperty(),
                    null, adaptivePropertiesDialog.selectedResourceProperty().getValue());
        }
    }

    @FXML
    private void onRefresh() {
        LOG.debug("onRefresh");
        Resource resource = adaptivePropertiesDialog.getSelectedResource();
        if (resource == null) {
            LOG.debug("no component_manager selected");
            return;
        }

        advertisementFetchManager.getAllResources(resource.getServer()).refreshAdvertisement();
    }

    public void updateAdvertisement(AdvertisementRspecSource advertisementRspecSource) {
        LOG.debug("Updating advertisement for bonfire " + (advertisementRspecSource == null
                ? "(there is no advertisement)"
                : "(there is an advertisement)"));

        if (advertisementRspecSource != null) {
            try {
                currentAdvertisement = (FXModelRspec) advertisementRspecSource.getModelRspec(ModelRspecType.FX);

                if (currentAdvertisement == null) {
                    clear();
                    return;
                }
            } catch (Exception e) {
                LOG.debug("Error parsing advertisment RSpec for bonfire", e);
                clear();
                return;
            }
        } else {
            clear();
            return;
        }

        assert currentAdvertisement != null;

        if (currentAdvertisement == null) {
            clear();
            return;
        }

        String prevLocation = locationComboBox.getSelectionModel().getSelectedItem();
        String prevSliverType = sliverTypeComboBox.getSelectionModel().getSelectedItem();

        locationIds.setAll(getLocationIds(currentAdvertisement));
        sliverTypes.setAll(getSliverTypes(currentAdvertisement));
        LOG.debug("   Set " + locationIds.size() + " locationIds");
        LOG.debug("   Set " + sliverTypes.size() + " sliverTypes");

        updateSpecificNode();

        if (locationComboBox.getSelectionModel().isEmpty() && prevLocation != null && locationIds.contains(prevLocation))
            locationComboBox.getSelectionModel().select(prevLocation);

        fallbackLocationToLoaded();

        if (locationComboBox.getSelectionModel().isEmpty() && !locationIds.isEmpty())
            locationComboBox.getSelectionModel().selectFirst();

        if (sliverTypeComboBox.getSelectionModel().isEmpty() && prevSliverType != null && sliverTypes.contains(prevSliverType))
            sliverTypeComboBox.getSelectionModel().select(prevSliverType);

        initForLocation(locationComboBox.getSelectionModel().getSelectedItem());

        if (sliverTypeComboBox.getSelectionModel().isEmpty() && prevSliverType != null && sliverTypes.contains(prevSliverType))
            sliverTypeComboBox.getSelectionModel().select(prevSliverType);

        fallbackOthersToLoaded();

        if (sliverTypeComboBox.getSelectionModel().isEmpty() && !sliverTypes.isEmpty())
            sliverTypeComboBox.getSelectionModel().selectFirst();
    }

    private void clear() {
        LOG.debug("clear()");
        currentAdvertisement = null;
        locationIds.clear();
        sliverTypes.clear();
        updateSpecificNode();

        //if nothing selected, but loaded values exist, use them
        fallbackToLoaded();
    }

    @FXML
    private void onLocationChange() {
        if (!locationComboBox.getSelectionModel().isEmpty())
            initForLocation(locationComboBox.getSelectionModel().getSelectedItem());
    }

    public void updateSpecificNode() {
        if (adaptivePropertiesDialog.getSelectedComponentId() != null) {
            GeniUrn ci = adaptivePropertiesDialog.getSelectedComponentId();
            assert ci != null;
            if (ci.getSubAuthName() != null) {
                locationComboBox.getSelectionModel().select(ci.getSubAuthName());
            } else
                LOG.warn("selected node has no sub auth: " + ci);
        }
    }

    public void initForLocation(String locationId) {
        LOG.debug("initForLocation(" + locationId + ")");
        if (currentAdvertisement == null) {
            locationId = null;
            LOG.debug("  abort: no current advertisement");
        }
        if (locationId == null) {
            this.lans.setOptions(Collections.<Lan>emptyList());
            this.diskImages.setOptions(Collections.<DiskImage>emptyList());
            lastInitializedLocation = null;
            LOG.debug("  abort initForLocation");
            return;
        }

        //only initialize if location differs from previous location
        if (locationId.equals(lastInitializedLocation)) {
            LOG.debug("  same initForLocation");
            return;
        }

        initLansForLocation(locationId);
        initDiskImagesForLocation(locationId);

        lastInitializedLocation = locationId;
    }

    public static Collection<String> getSliverTypes(ModelRspec advertisementModelRspec) {
        if (advertisementModelRspec.getNodes() == null)
            return Collections.emptyList();
        return advertisementModelRspec.getNodes().stream()
                .flatMap(node -> node.getSliverTypes().stream())
                .map(SliverType::getName)
                .filter(Objects::nonNull)
                .distinct()
                .sorted()
                .collect(Collectors.toList());
    }

    public static Collection<String> getLocationIds(ModelRspec advertisementModelRspec) {
        if (advertisementModelRspec.getNodes() == null)
            return Collections.emptyList();
        return advertisementModelRspec.getNodes().stream().map(RspecNode::getComponentManagerId)
                .filter(Objects::nonNull)
                .map(GeniUrn::getSubAuthName)
                .filter(Objects::nonNull)
                .distinct()
                .sorted()
                .collect(Collectors.toList());

    }

    private void initLansForLocation(String locationId) {
        Set<String> prevSelectedNetworks = new HashSet<>();

        if (currentLocationBonfireNodeInfo != null) {
            prevSelectedNetworks.addAll(Lists.transform(currentLocationBonfireNodeInfo.getNetworkInfos(),
                    BonfireNodeInfo.NetworkInfo::getHref
            ));

        } else if (!this.lans.getSelectedItems().isEmpty()) {
            prevSelectedNetworks.addAll(Lists.transform(this.lans.getSelectedItems(),
                    Lan::getComponentId
            ));
        }
        LOG.debug("  prevSelectedNetworks.size=" + prevSelectedNetworks.size());

        boolean autoSelectNetwork =
                !Objects.equals(lastLocationForWhichDefaultNetworkSelected, locationId) && prevSelectedNetworks.isEmpty();
        LOG.debug("initForLocation locationId={} autoSelectNetwork={} prevSelectedNetworks.size()={}",
                locationId, autoSelectNetwork, prevSelectedNetworks.size());

        List<Lan> newLans = new ArrayList<>();
        for (Lan lan : currentAdvertisement.getBonfireLans()) {
            GeniUrn cmi = GeniUrn.parse(lan.getComponentManagerId());
            if (cmi == null) continue;
            String subAuthName = cmi.getSubAuthName();
            if (subAuthName == null) continue;
            if (locationId.equals(subAuthName)) {
                newLans.add(lan);
                LOG.debug("initForLocation adding lan: " + lan.getComponentName());
                if (autoSelectNetwork && lan.getComponentName().toLowerCase().contains("WAN")) {
                    prevSelectedNetworks.add(lan.getComponentId());
                    autoSelectNetwork = false;
                }
            }
        }
        this.lans.setOptions(newLans);
        LOG.debug("  newLans.size=" + newLans.size());

        if (!newLans.isEmpty()) {
            List<Lan> selectedItems =
                    newLans.stream().filter(lan -> prevSelectedNetworks.contains(lan.getComponentId())).collect(Collectors.toList());
            this.lans.setSelectedOptions(selectedItems);
            lastLocationForWhichDefaultNetworkSelected = locationId;

            this.lans.setDisable(false);
        }
    }

    public void initDiskImagesForLocation(String locationId) {
        if (locationId == null) {
            this.diskImages.setOptions(Collections.<DiskImage>emptyList());
            return;
        }

        List<DiskImage> newDiskImages = new ArrayList<>();
        Set<String> prevSelectedDiskImages = new HashSet<>();

        if (currentLocationBonfireNodeInfo != null) {
            prevSelectedDiskImages.addAll(Lists.transform(currentLocationBonfireNodeInfo.getStorageInfos(),
                    BonfireNodeInfo.StorageInfo::getHref
            ));

        } else if (!this.lans.getSelectedItems().isEmpty()) {
            prevSelectedDiskImages.addAll(Lists.transform(this.diskImages.getSelectedItems(),
                    DiskImage::getComponentId
            ));
        }

        boolean autoSelectStorage =
                !Objects.equals(lastLocationForWhichDefaultStorageSelected, locationId) && prevSelectedDiskImages.isEmpty();
        for (DiskImage diskImage : currentAdvertisement.getBonfireDiskImages()) {
            GeniUrn cmi = GeniUrn.parse(diskImage.getComponentManagerId());
            if (cmi == null) continue;
            String subAuthName = cmi.getSubAuthName();
            if (subAuthName == null) continue;
            if (locationId.equals(subAuthName)) {
                newDiskImages.add(diskImage);
                LOG.debug("initForLocation dding storage: " + diskImage.getComponentName());

                if (autoSelectStorage) {
                    if (diskImage.getComponentName().equalsIgnoreCase("BonFIRE Debian Squeeze v6")) {
                        prevSelectedDiskImages.add(diskImage.getComponentId());
                        autoSelectStorage = false;
                    }
                }
            }
        }
        this.diskImages.setOptions(newDiskImages);

        if (!newDiskImages.isEmpty()) {
            List<DiskImage> selectedItems =
                    newDiskImages.stream().filter(storage -> prevSelectedDiskImages.contains(storage.getComponentId())).collect(Collectors.toList());
            this.diskImages.setSelectedOptions(selectedItems);
            lastLocationForWhichDefaultStorageSelected = locationId;

            this.diskImages.setDisable(false);
        }
    }

    @FXML
    private void onSaveAction(ActionEvent actionEvent) {
        adaptivePropertiesDialog.onSaveAction(actionEvent);
    }

    String loadedLocation;
    String loadedSliverType;
    List<DiskImage> loadedDiskImages;
    List<Lan> loadedLans;

    @Override
    public void load(FXRspecNode rspecNode) {
        if (rspecNode == null || !isBonfire(rspecNode.getComponentManagerId()))
            return;

        loadedLocation = rspecNode.getComponentManagerId().getSubAuthName();
        loadedSliverType = rspecNode.getSliverTypeName() != null ? rspecNode.getSliverTypeName().trim() : null;
        loadedDiskImages = rspecNode.getBonfireDiskImages();
        loadedLans = rspecNode.getBonfireLans();

        fallbackToLoaded();
    }

    private void fallbackToLoaded() {
        fallbackLocationToLoaded();
        fallbackOthersToLoaded();
    }

    private void fallbackLocationToLoaded() {
        if (loadedLocation != null && !loadedLocation.trim().isEmpty()) {
            if (!locationComboBox.getItems().contains(loadedLocation))
                locationComboBox.getItems().add(loadedLocation);
            locationComboBox.getSelectionModel().select(loadedLocation);
        }
    }

    private void fallbackOthersToLoaded() {
        if (loadedSliverType != null && !loadedSliverType.trim().isEmpty()) {
            if (sliverTypeComboBox.getItems().isEmpty())
                sliverTypeComboBox.getItems().add(loadedSliverType);
            sliverTypeComboBox.getSelectionModel().select(loadedSliverType);
        }

        if (loadedDiskImages != null && !loadedDiskImages.isEmpty()) {
            if (diskImages.isEmpty())
                diskImages.setOptions(loadedDiskImages);
            if (diskImages.isEmptySelection())
                diskImages.setSelectedOptions(loadedDiskImages);
        }

        if (loadedLans != null && !loadedLans.isEmpty()) {
            if (lans.isEmpty())
                lans.setOptions(loadedLans);
            if (lans.isEmptySelection())
                lans.setSelectedOptions(loadedLans);
        }
    }

    @Override
    public void save(FXRspecNode rspecNode) {
        assert rspecNode != null;

        String selectedTypeName = sliverTypeComboBox.getSelectionModel().getSelectedItem();
        if (selectedTypeName != null)
            rspecNode.setSliverTypeName(selectedTypeName);

        if (!hasFixedNode.get()) {
            String locationId = locationComboBox.getSelectionModel().getSelectedItem();
            GeniUrn cmi = adaptivePropertiesDialog.getSelectedResource().getServer().getDefaultComponentManagerAsGeniUrn();
            rspecNode.setComponentManagerId(GeniUrn.createGeniUrnFromEncodedParts(
                    cmi.getEncodedTopLevelAuthority_withoutSubAuth() + ":" + locationId,
                    cmi.getResourceType(),
                    cmi.getResourceName()
            ));
        }

        //save lans and diskimages
        List<DiskImage> selectedDiskImages = new ArrayList<>(diskImages.getSelectedItems());
        List<Lan> selectedLans = new ArrayList<>(lans.getSelectedItems());
        rspecNode.setBonfireDiskImages(selectedDiskImages);
        rspecNode.setBonfireLans(selectedLans);
    }

    @Override
    public void onTabShown() {
        currentResource =
                adaptivePropertiesDialog.getSelectedResource();
        assert isBonfire(currentResource.getServer().getDefaultComponentManagerAsGeniUrn());

        ReadOnlyObjectProperty<AdvertisementRspecSource> anyAdvProp =
                                    config.getAdvertisement(currentResource.getServer(), Optional.<Boolean>empty()/*= any advertisement*/);

        if (anyAdvProp.get() == null)
            onRefresh(); //force refresh
    }

    @Override
    public boolean shouldBeVisible(FXRspecNode rspecNode) {
        if (adaptivePropertiesDialog.getSelectedResource() == null) return false;
        if (adaptivePropertiesDialog.getSelectedResource().getServer() == null) return false;
        return isBonfire(adaptivePropertiesDialog.getSelectedResource().getServer().getDefaultComponentManagerAsGeniUrn());
    }

    @Override
    public String validateConfiguration() {
        if (locationComboBox.getSelectionModel().isEmpty()) {
            return "The location of a Bonfire Node must be specified";
        }
        if (sliverTypeComboBox.getSelectionModel().isEmpty()) {
            return "The instance type (= sliver_type) of a Bonfire Node must be specified";
        }

        return null;
    }

    public static boolean isBonfire(GeniUrn urn) {
        if (urn == null) return false;
        return urn.getEncodedTopLevelAuthority_withoutSubAuth().endsWith("bonfire.eu");
    }
}
